#
# (C) Copyright 2018 InovaDevelopment.com
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
Test for the WBEMServer class  in pywbem._server.py that uses the pywbem_mock
support package the methods of the class. Mock is required since, testing
WBEMServer requires access to a WBEM server.
"""
from __future__ import absolute_import, print_function

import os
import pytest

from pywbem import ValueMapping, CIMInstanceName, CIMError, \
    CIMQualifierDeclaration, CIMClass
from pywbem._nocasedict import NocaseDict

from ..utils.wbemserver_mock import WbemServerMock
from ..utils.pytest_extensions import simplified_test_function

VERBOSE = True

# Location of DMTF schema directory used by all tests.
# This directory is permanent and should not be removed.
TESTSUITE_SCHEMA_DIR = os.path.join('tests', 'schema')


class BaseMethodsForTests(object):
    """
    Common methods for test of WBEMServer class.  This includes methods to
    build the DMTF schema and to build individual instances.
    """


class TestServerClass(BaseMethodsForTests):
    # pylint: disable=too-few-public-methods
    """
    Conduct tests on the WBEMServer class.
    """

    @pytest.mark.parametrize(
        "tst_namespace",
        ['interop', 'root/interop', 'root/PG_Interop'])
    def test_wbemserver_basic(self, tst_namespace):
        # pylint: disable=no-self-use
        """
        Test the basic functions that access server information. This test
        creates the mock repository and adds classes and instances for
        the WBEMServer tests that involve namespaces, brand, profiles and
        a subset of the central_instance tests.  It includes no tests for
        errors. The primary goal of this test was to develop the mechanisms
        for easily getting classes and instances into the repo and to provide
        a basic test of functionality.
        """

        # Build the wbem server mock using the  WbemServerMock default test
        # data except that we define the interop namespace
        mock_wbemserver = WbemServerMock(interop_ns=tst_namespace)
        server = mock_wbemserver.wbem_server

        # Build instances for get_central instance
        # Using central methodology, i.e. ElementConformsToProfile

        # Test basic brand, version, namespace methods
        assert server.namespace_classname == 'CIM_Namespace'

        assert server.url == 'http://FakedUrl'

        assert server.brand == "OpenPegasus"
        assert server.version == "2.15.0"
        assert server.interop_ns == tst_namespace
        assert set(server.namespaces) == set([tst_namespace])

        # Test basic profiles methods
        org_vm = ValueMapping.for_property(server, server.interop_ns,
                                           'CIM_RegisteredProfile',
                                           'RegisteredOrganization')

        for inst in server.profiles:
            org = org_vm.tovalues(inst['RegisteredOrganization'])
            name = inst['RegisteredName']
            vers = inst['RegisteredVersion']

            tst_tup = (org, name, vers)
            pass_tst = False
            for tup in mock_wbemserver.registered_profiles:
                if tst_tup == tup:
                    pass_tst = True
                    break
            assert pass_tst

        sel_prof = server.get_selected_profiles(registered_org='DMTF',
                                                registered_name='Indications')
        assert len(sel_prof) == 1
        for inst in sel_prof:
            assert org_vm.tovalues(inst['RegisteredOrganization']) == 'DMTF'
            assert inst['RegisteredName'] == 'Indications'

        # Test case insensitive matching
        sel_prof = server.get_selected_profiles(registered_org='DmtF',
                                                registered_name='inDiCations')
        assert len(sel_prof) == 1
        for inst in sel_prof:
            assert org_vm.tovalues(inst['RegisteredOrganization']) == 'DMTF'
            assert inst['RegisteredName'] == 'Indications'

        sel_prof = server.get_selected_profiles(registered_org='DMTF')
        assert len(sel_prof) == 3
        for inst in sel_prof:
            assert org_vm.tovalues(inst['RegisteredOrganization']) == 'DMTF'

        # Simple get_cental_instance.
        # profile_path, central_class=None,
        #                       scoping_class=None, scoping_path=None
        profile_insts = server.get_selected_profiles(registered_org='SNIA',
                                                     registered_name='Server',
                                                     registered_version='1.1.0')
        profile_path = profile_insts[0].path
        insts = server.get_central_instances(profile_path, 'CIM_ObjectManager')
        assert len(insts) == 1
        kb = NocaseDict([('SystemCreationClassName', 'CIM_ComputerSystem'),
                         ('SystemName', mock_wbemserver.system_name),
                         ('CreationClassName', 'CIM_ObjectManager'),
                         ('Name', 'MyFakeObjectManager'), ])
        assert insts[0] == CIMInstanceName('CIM_ObjectManager', keybindings=kb,
                                           namespace=tst_namespace,
                                           host=server.conn.host)


TESTCASES_CREATE_NAMESPACE = [

    # Testcases for WBEMServer.create_namespace()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * new_namespace: Name of the namespace to be created.
    #   * exp_namespace: Expected returned namespace name.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "New top level namespace",
        dict(
            new_namespace='abc',
            exp_namespace=u'abc',
        ),
        None, None, True
    ),
    (
        "New top level namespace with leading and trailing slash",
        dict(
            new_namespace='/abc/',
            exp_namespace=u'abc',
        ),
        None, None, True
    ),
    (
        "New two-segment namespace with leading and trailing slash",
        dict(
            new_namespace='/abc/def/',
            exp_namespace=u'abc/def',
        ),
        None, None, True
    ),
    (
        "New two-segment namespace where first segment already exists",
        dict(
            new_namespace='interop/def',
            exp_namespace=u'interop/def',
        ),
        None, None, True
    ),
    (
        "Existing top level namespace",
        dict(
            new_namespace='interop',
            exp_namespace=None,
        ),
        CIMError, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_CREATE_NAMESPACE)
@simplified_test_function
def test_create_namespace_2(testcase,
                            new_namespace, exp_namespace):
    """
    Test creation of a namespace using approach 2.
    """

    mock_wbemserver = WbemServerMock(interop_ns='interop')
    server = mock_wbemserver.wbem_server

    act_namespace = server.create_namespace(new_namespace)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_namespace == exp_namespace


TESTCASES_DELETE_NAMESPACE = [

    # Testcases for WBEMServer.delete_namespace()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * namespace: Name of the namespace to be deleted.
    #   * namespace_content: Content of initial namespace, as a dict with
    #     the namespace name as a key and a list of CIMClass, CIMInstance and
    #     CIMQualifierDeclaration objects as the value.
    #   * exp_namespace: Expected returned namespace name.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty top level namespace",
        dict(
            namespace='abc',
            namespace_content={'abc': []},
            exp_namespace=u'abc',
        ),
        None, None, True
    ),
    (
        "Empty top level namespace with leading and trailing slash",
        dict(
            namespace='/abc/',
            namespace_content={'abc': []},
            exp_namespace=u'abc',
        ),
        None, None, True
    ),
    (
        "Empty two-segment namespace with leading and trailing slash",
        dict(
            namespace='/abc/def/',
            namespace_content={'abc/def': []},
            exp_namespace=u'abc/def',
        ),
        None, None, True
    ),
    (
        "Empty two-segment namespace where first segment already exists",
        dict(
            namespace='interop/def',
            namespace_content={'interop/def': []},
            exp_namespace=u'interop/def',
        ),
        None, None, True
    ),
    (
        "Non-existing top level namespace",
        dict(
            namespace='abc',
            namespace_content={},
            exp_namespace=None,
        ),
        CIMError, None, True
    ),
    (
        "Non-empty top level namespace containing a class",
        dict(
            namespace='abc',
            namespace_content={'abc': [
                CIMClass('Foo')
            ]},
            exp_namespace=None,
        ),
        CIMError, None, True
    ),
    (
        "Non-empty top level namespace containing a qualifier type",
        dict(
            namespace='abc',
            namespace_content={'abc': [
                CIMQualifierDeclaration('Foo', 'string')
            ]},
            exp_namespace=None,
        ),
        CIMError, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_DELETE_NAMESPACE)
@simplified_test_function
def test_delete_namespace(testcase,
                          namespace, namespace_content, exp_namespace):
    """
    Test deletion of a namespace.
    """

    mock_wbemserver = WbemServerMock(interop_ns='interop')
    server = mock_wbemserver.wbem_server

    # Ensure that the namespace is set up as specified in the testcase
    for ns in namespace_content:
        server.create_namespace(ns)
        for obj in namespace_content[ns]:
            if isinstance(obj, CIMClass):
                server.conn.CreateClass(NewClass=obj, namespace=ns)
            else:
                assert isinstance(obj, CIMQualifierDeclaration)
                server.conn.SetQualifier(QualifierDeclaration=obj,
                                         namespace=ns)

    # The code to be tested
    act_namespace = server.delete_namespace(namespace)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_namespace == exp_namespace


TESTCASES_GET_CENTRAL_INSTANCES = [

    # Testcases for WBEMServer.get_central_instances()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * profile_name: Name of the profile as tuple (org, name, version)
    #   * central_class: Name of the central class of the profile.
    #   * scoping_class: Name of the scoping class of the profile, or None for
    #     autonomous profiles.
    #   * scoping_path: List of class names for the scoping path (from central
    #     class to scoping class, not including central or scoping classes), or
    #     None for autonomous profiles.
    #   * direction: Reference direction 'snia' or 'dmtf'.
    #   * exp_paths: Expected returned instance paths, as a list of
    #     CIMInstanceName objects.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger
    (
        "Valid central class",
        dict(
            profile_name=('SNIA', 'Server', '1.2.0'),
            central_class='XXX_StorageComputerSystem',
            scoping_class=None,
            scoping_path=None,
            direction='snia',
            exp_paths=[
                CIMInstanceName('XXX_StorageComputerSystem',
                                keybindings={'Name': "10.1.2.3",
                                             'CreationClassName':
                                             'XXX_StorageComputerSystem'},
                                host='FakedUrl',
                                namespace='interop'), ]
        ),
        None, None, True
    ),
    # TODO add more central instance tests
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_GET_CENTRAL_INSTANCES)
@simplified_test_function
def test_get_central_instances(testcase, profile_name, central_class,
                               scoping_class, scoping_path, direction,
                               exp_paths):
    """
    Test the execution of WBEMSeever.get_central_instances()
    """
    mock_wbemserver = WbemServerMock(interop_ns='interop')
    server = mock_wbemserver.wbem_server
    profile_insts = server.get_selected_profiles(
        registered_org=profile_name[0],
        registered_name=profile_name[1],
        registered_version=profile_name[2])
    assert len(profile_insts) == 1
    profile_inst = profile_insts[0]

    # The code to be tested.
    # Must start from a single good profile instance

    # server.conn.display_repository()
    act_paths = server.get_central_instances(profile_inst.path,
                                             central_class=central_class,
                                             scoping_class=scoping_class,
                                             scoping_path=scoping_path,
                                             reference_direction=direction)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_paths == exp_paths


# TODO Break up tests to do individual tests for each group of methods so we can
#      test for errors, variations on what is in the repo with each method.
#      Right now we build it all in a single test.  Thus, for example we
#      need to create a test group for find_central_instances since the
#      definition of the repo is different for each method of getting the
#      central instances Iex. If the server method exists, no other methods
#      are tried.



"""
Test the _nocasedict module.
"""

from __future__ import absolute_import

import sys
import re
import six
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import pytest

from pywbem._nocasedict import NocaseDict
from pywbem._utils import _format

from ..utils.pytest_extensions import simplified_test_function


class NonComparable(object):
    # pylint: disable=too-few-public-methods
    """
    Class that raises TypeError when comparing its objects for equality or
    when hashing its objects.
    """

    def __eq__(self, other):
        raise TypeError("Cannot compare %s to %s" % (type(self), type(other)))

    def __ne__(self, other):
        raise TypeError("Cannot compare %s to %s" % (type(self), type(other)))

    def __hash__(self):
        raise TypeError("Cannot hash %s" % type(self))


TESTCASES_NOCASEDICT_INIT = [

    # Testcases for NocaseDict.__init__() / ncd=NocaseDict()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to NocaseDict().
    #   * init_kwargs: Dict of keyword arguments to NocaseDict().
    #   * exp_dict: Expected resulting dictionary, as OrderedDict.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict from no args",
        dict(
            init_args=(),
            init_kwargs={},
            exp_dict=OrderedDict(),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty dict from None as positional arg",
        dict(
            init_args=(None,),
            init_kwargs={},
            exp_dict=OrderedDict(),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty dict from empty list as positional arg",
        dict(
            init_args=(list(),),
            init_kwargs={},
            exp_dict=OrderedDict(),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty dict from empty tuple as positional arg",
        dict(
            init_args=(tuple(),),
            init_kwargs={},
            exp_dict=OrderedDict(),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty dict from empty dict as positional arg",
        dict(
            init_args=(dict(),),
            init_kwargs={},
            exp_dict=OrderedDict(),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Empty dict from empty NocaseDict as positional arg",
        dict(
            init_args=(NocaseDict(),),
            init_kwargs={},
            exp_dict=OrderedDict(),
            verify_order=True,
        ),
        None, None, True
    ),

    # Non-empty NocaseDict
    (
        "Dict from list as positional arg",
        dict(
            init_args=([('Dog', 'Cat'), ('Budgie', 'Fish')],),
            init_kwargs={},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Dict from tuple as positional arg",
        dict(
            init_args=((('Dog', 'Cat'), ('Budgie', 'Fish')),),
            init_kwargs={},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Dict from dict as positional arg",
        dict(
            init_args=({'Dog': 'Cat', 'Budgie': 'Fish'},),
            init_kwargs={},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=False,
        ),
        None, UserWarning, True
    ),
    (
        "Dict from keyword args",
        dict(
            init_args=(),
            init_kwargs={'Dog': 'Cat', 'Budgie': 'Fish'},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=False,
        ),
        None, UserWarning if sys.version_info[0:2] < (3, 7) else None, True
    ),
    (
        "Dict from list as positional arg and keyword args",
        dict(
            init_args=([('Dog', 'Cat')],),
            init_kwargs={'Budgie': 'Fish'},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Dict from tuple as positional arg and keyword args",
        dict(
            init_args=((('Dog', 'Cat'),),),
            init_kwargs={'Budgie': 'Fish'},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=True,
        ),
        None, None, True
    ),
    (
        "Dict from dict as positional arg and keyword args",
        dict(
            init_args=({'Dog': 'Cat'},),
            init_kwargs={'Budgie': 'Fish'},
            exp_dict=OrderedDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            verify_order=True,
        ),
        None, None, True
    ),

    # Error cases
    (
        "String type as positional arg",
        dict(
            init_args=('illegal',),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
    (
        "Two positional args",
        dict(
            init_args=(list(), list()),
            init_kwargs={},
            exp_dict=None,
            verify_order=None,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_INIT)
@simplified_test_function
def test_NocaseDict_init(testcase,
                         init_args, init_kwargs, exp_dict, verify_order):
    """
    Test function for NocaseDict.__init__() / ncd=NocaseDict()
    """

    # The code to be tested
    act_dict = NocaseDict(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseDict features, but that is
    # unavoidable if we want to work through the public interface:

    act_items = []
    for key in act_dict:  # Uses NocaseDict iteration
        act_value = act_dict[key]  # Uses NocaseDict getitem
        assert key in exp_dict, "Unexpected extra key %r" % key
        exp_value = exp_dict[key]
        assert act_value == exp_value, "Unexpected value at key %r" % key
        act_items.append((key, act_value))

    exp_items = []
    for key in exp_dict:
        exp_value = exp_dict[key]
        # Next line uses NocaseDict contains:
        assert key in act_dict, "Unexpected missing key %r" % key
        act_value = act_dict[key]  # Uses NocaseDict getitem
        assert act_value == exp_value, "Unexpected value at key %r" % key
        exp_items.append((key, exp_value))

    if verify_order:
        assert act_items == exp_items, "Unexpected order of items"


TESTCASES_NOCASEDICT_GETITEM = [

    # Testcases for NocaseDict.__getitem__() / ncd[key]

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * key: Key to be used for the test.
    #   * exp_value: Expected value for the key.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=1234,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key (not found)",
        dict(
            obj=NocaseDict(),
            key='',
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Empty dict, with non-empty key (not found)",
        dict(
            obj=NocaseDict(),
            key='Dog',
            exp_value=None,
        ),
        KeyError, None, True
    ),

    # Non-empty NocaseDict
    (
        "Non-empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with empty string key (not found)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (not found)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
            exp_value=None,
        ),
        KeyError, None, True
    ),
    (
        "Non-empty dict, with existing key in original case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original upper case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='DOG',
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original lower case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='dog',
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original mixed case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            exp_value='Cat',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_GETITEM)
@simplified_test_function
def test_NocaseDict_getitem(testcase,
                            obj, key, exp_value):
    """
    Test function for NocaseDict.__getitem__() / ncd[key]
    """

    # The code to be tested
    act_value = obj[key]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_value == exp_value, "Unexpected value at key %r" % key


TESTCASES_NOCASEDICT_SETITEM = [

    # Testcases for NocaseDict.__setitem__() / ncd[key]=value

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * key: Key to be used for the test.
    #   * value: New value and expected value to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=None,
            value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=1234,
            value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key",
        dict(
            obj=NocaseDict(),
            key='',
            value='Newbie',
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key",
        dict(
            obj=NocaseDict(),
            key='Dog',
            value='Kitten',
        ),
        None, None, True
    ),

    # Non-empty NocaseDict
    (
        "Non-empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key=None,
            value='Kitten',
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with empty string key",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='newkey',
            value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in original case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            value='Kitten',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original upper case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='DOG',
            value='Kitten',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original lower case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='dog',
            value='Kitten',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original mixed case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            value='Kitten',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_SETITEM)
@simplified_test_function
def test_NocaseDict_setitem(testcase, obj, key, value):
    """
    Test function for NocaseDict.__setitem__() / ncd[key]=value
    """

    # The code to be tested
    obj[key] = value

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseDict features, but that is
    # unavoidable if we want to work through the public interface:

    act_value = obj[key]  # Uses NocaseDIct getitem

    assert act_value == value, "Unexpected value at key %r" % key


TESTCASES_NOCASEDICT_DELITEM = [

    # Testcases for NocaseDict.__delitem__() / del ncd[key]

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * key: Key to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=1234,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key (not found)",
        dict(
            obj=NocaseDict(),
            key='',
        ),
        KeyError, None, True
    ),
    (
        "Empty dict, with non-empty key (not found)",
        dict(
            obj=NocaseDict(),
            key='Dog',
        ),
        KeyError, None, True
    ),

    # Non-empty NocaseDict
    (
        "Non-empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with empty non-existing string key (not found)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
        ),
        KeyError, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (not found)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
        ),
        KeyError, None, True
    ),
    (
        "Non-empty dict, with existing key in original case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original upper case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='DOG',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original lower case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='dog',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original mixed case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_DELITEM)
@simplified_test_function
def test_NocaseDict_delitem(testcase, obj, key):
    """
    Test function for NocaseDict.__delitem__() / del ncd[key]
    """

    # The code to be tested
    del obj[key]

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseDict features, but that is
    # unavoidable if we want to work through the public interface:

    with pytest.raises(KeyError):
        # The following line uses NocaseDict getitem
        obj[key]  # pylint: disable=pointless-statement


TESTCASES_NOCASEDICT_LEN = [

    # Testcases for NocaseDict.__len__() / len(ncd)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * exp_len: Expected len() value.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=NocaseDict(),
            exp_len=0,
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            exp_len=2,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_LEN)
@simplified_test_function
def test_NocaseDict_len(testcase, obj, exp_len):
    """
    Test function for NocaseDict.__len__() / len(ncd)
    """

    # The code to be tested
    act_len = len(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_len == exp_len


TESTCASES_NOCASEDICT_CONTAINS = [

    # Testcases for NocaseDict.__contains__() / key in ncd

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * key: Key to be used for the test.
    #   * exp_result: Expected result (bool).
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=None,
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=1234,
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key (not found)",
        dict(
            obj=NocaseDict(),
            key='',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (not found)",
        dict(
            obj=NocaseDict(),
            key='Dog',
            exp_result=False,
        ),
        None, None, True
    ),

    # Non-empty NocaseDict
    (
        "Non-empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key=None,
            exp_result=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with empty non-existing string key (not found)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (not found)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
            exp_result=False,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in original case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            exp_result=True,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original upper case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='DOG',
            exp_result=True,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original lower case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='dog',
            exp_result=True,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in non-original mixed case",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            exp_result=True,
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_CONTAINS)
@simplified_test_function
def test_NocaseDict_contains(testcase, obj, key, exp_result):
    """
    Test function for NocaseDict.__contains__() / key in ncd
    """

    # The code to be tested
    act_result = key in obj

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_result == exp_result, "Unexpected result at key %r" % key


TESTCASES_NOCASEDICT_GET = [

    # Testcases for NocaseDict.get()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * key: Key to be used for the test.
    #   * default: Default value to be used for the test, or None to not pass.
    #   * exp_value: Expected value at the key.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=None,
            default=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=1234,
            default=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key (defaulted without default)",
        dict(
            obj=NocaseDict(),
            key='',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key (defaulted to a value)",
        dict(
            obj=NocaseDict(),
            key='',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (defaulted without default)",
        dict(
            obj=NocaseDict(),
            key='Dog',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (defaulted to a value)",
        dict(
            obj=NocaseDict(),
            key='Dog',
            default='Kitten',
            exp_value='Kitten',
        ),
        None, None, True
    ),

    # Non-empty NocaseDict
    (
        "Non-empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key=None,
            default=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with empty string key (defaulted without default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with empty string key (defaulted to a value)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (defaulted without "
        "default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (defaulted to a "
        "value)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in original case (no default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            default=None,
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in original case (with default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            default='Newbie',
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in mixed case (no default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            default=None,
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in mixed case (with default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            default='Newbie',
            exp_value='Cat',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_GET)
@simplified_test_function
def test_NocaseDict_get(testcase, obj, key, default, exp_value):
    """
    Test function for NocaseDict.get()
    """

    # The code to be tested
    if default is None:
        act_value = obj.get(key)
    else:
        act_value = obj.get(key, default)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_value == exp_value, "Unexpected value at key %r with " \
                                   "default %r" % (key, default)


TESTCASES_NOCASEDICT_SETDEFAULT = [

    # Testcases for NocaseDict.setdefault()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * key: Key to be used for the test.
    #   * default: Default value to be used for the test (None is passed).
    #   * exp_value: Expected value at the key.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=None,
            default=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key (invalid type)",
        dict(
            obj=NocaseDict(),
            key=1234,
            default=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key (defaulted without default)",
        dict(
            obj=NocaseDict(),
            key='',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key (defaulted to a value)",
        dict(
            obj=NocaseDict(),
            key='',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (defaulted without default)",
        dict(
            obj=NocaseDict(),
            key='Dog',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty key (defaulted to a value)",
        dict(
            obj=NocaseDict(),
            key='Dog',
            default='Kitten',
            exp_value='Kitten',
        ),
        None, None, True
    ),

    # Non-empty NocaseDict
    (
        "Non-empty dict, with None key (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key=None,
            default=None,
            exp_value=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with empty string key (defaulted without default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with empty string key (defaulted to a value)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (defaulted without "
        "default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
            default=None,
            exp_value=None,
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with non-empty non-existing key (defaulted to a "
        "value)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='invalid',
            default='Newbie',
            exp_value='Newbie',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in original case (no default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            default=None,
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in original case (with default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='Dog',
            default='Newbie',
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in mixed case (no default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            default=None,
            exp_value='Cat',
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with existing key in mixed case (with default)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            key='doG',
            default='Newbie',
            exp_value='Cat',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_SETDEFAULT)
@simplified_test_function
def test_NocaseDict_setdefault(testcase, obj, key, default, exp_value):
    """
    Test function for NocaseDict.setdefault()
    """

    # The code to be tested
    act_value = obj.setdefault(key, default)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_value == exp_value, "Unexpected value at key %r with " \
                                   "default %r" % (key, default)


TESTCASES_NOCASEDICT_ITEMS = [

    # Testcases for NocaseDict.keys(), values(), items()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * exp_items: List with expected items (key,value) in expected order.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=NocaseDict(),
            exp_items=[],
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            exp_items=[('Dog', 'Cat'), ('Budgie', 'Fish')],
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_keys(testcase, obj, exp_items):
    """
    Test function for NocaseDict.keys()
    """

    # The code to be tested
    act_keys = obj.keys()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_keys = [item[0] for item in exp_items]
    assert act_keys == exp_keys


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_values(testcase, obj, exp_items):
    """
    Test function for NocaseDict.values()
    """

    # The code to be tested
    act_values = obj.values()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_values = [item[1] for item in exp_items]
    assert act_values == exp_values


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_items(testcase, obj, exp_items):
    """
    Test function for NocaseDict.items()
    """

    # The code to be tested
    act_items = obj.items()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_items == exp_items


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_iterkeys(testcase, obj, exp_items):
    """
    Test function for NocaseDict.iterkeys()
    """

    # The code to be tested
    act_keys = []
    for key in obj.iterkeys():
        act_keys.append(key)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_keys = [item[0] for item in exp_items]
    assert act_keys == exp_keys


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_itervalues(testcase, obj, exp_items):
    """
    Test function for NocaseDict.itervalues()
    """

    # The code to be tested
    act_values = []
    for value in obj.itervalues():
        act_values.append(value)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    exp_values = [item[1] for item in exp_items]
    assert act_values == exp_values


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_iteritems(testcase, obj, exp_items):
    """
    Test function for NocaseDict.iteritemss()
    """

    # The code to be tested
    act_items = []
    for item in obj.iteritems():
        act_items.append(item)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_items == exp_items


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ITEMS)
@simplified_test_function
def test_NocaseDict_iter(testcase, obj, exp_items):
    """
    Test function for NocaseDict.__iter__() / for key in ncd
    """

    # The code to be tested
    act_items = []
    for key in obj:
        value = obj[key]
        act_items.append((key, value))

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert act_items == exp_items


TESTCASES_NOCASEDICT_REPR = [

    # Testcases for NocaseDict.__repr__() / repr(ncd)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=NocaseDict(),
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_REPR)
@simplified_test_function
def test_NocaseDict_repr(testcase, obj):
    """
    Test function for NocaseDict.__repr__() / repr(ncd)
    """

    # The code to be tested
    result = repr(obj)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert re.match(r'^NocaseDict\(.*\)$', result)

    # Note: This only tests for existence of each item, not for excess items
    # or representing the correct order.
    for item in obj.items():
        exp_item_result = _format("{0!A}: {1!A}", *item)
        assert exp_item_result in result


TESTCASES_NOCASEDICT_UPDATE = [

    # Testcases for NocaseDict.update()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * args: List of positional args for update().
    #   * kwargs: Dict of keyword args for update().
    #   * exp_obj: Expected NocaseDict after being updated.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty NocaseDict
    (
        "Empty dict, with empty update args + kwargs",
        dict(
            obj=NocaseDict(),
            args=[],
            kwargs={},
            exp_obj=NocaseDict(),
        ),
        None, None, True
    ),
    (
        "Empty dict, with integer key in update args (invalid type)",
        dict(
            obj=NocaseDict(),
            args=[[(1234, 'Invalid')]],
            kwargs={},
            exp_obj=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with integer key in update kwargs (invalid type)",
        dict(
            obj=NocaseDict(),
            args=[],
            kwargs={1234: 'Invalid'},
            exp_obj=None,
        ),
        TypeError, None, True
    ),
    (
        "Empty dict, with empty string key in update args+items",
        dict(
            obj=NocaseDict(),
            args=[OrderedDict([('', 'Cat')])],
            kwargs={},
            exp_obj=NocaseDict([('', 'Cat')]),
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key in update args+iter",
        dict(
            obj=NocaseDict(),
            args=[[('', 'Cat')]],
            kwargs={},
            exp_obj=NocaseDict([('', 'Cat')]),
        ),
        None, None, True
    ),
    (
        "Empty dict, with empty string key in update kwargs",
        dict(
            obj=NocaseDict(),
            args=[],
            kwargs={'': 'Cat'},
            exp_obj=NocaseDict([('', 'Cat')]),
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty string key in update args+items",
        dict(
            obj=NocaseDict(),
            args=[OrderedDict([('Dog', 'Cat')])],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Cat')]),
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty string key in update args+iter",
        dict(
            obj=NocaseDict(),
            args=[[('Dog', 'Cat')]],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Cat')]),
        ),
        None, None, True
    ),
    (
        "Empty dict, with non-empty string key in update kwargs",
        dict(
            obj=NocaseDict(),
            args=[],
            kwargs={'Dog': 'Cat'},
            exp_obj=NocaseDict([('Dog', 'Cat')]),
        ),
        None, None, True
    ),

    # Non-empty NocaseDict, insert new value
    (
        "Non-empty dict, with integer key in update args (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[[(1234, 'Invalud')]],
            kwargs={},
            exp_obj=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with integer key in update kwargs (invalid type)",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[],
            kwargs={1234: 'Invalid'},
            exp_obj=None,
        ),
        TypeError, None, True
    ),
    (
        "Non-empty dict, with new empty string key in update args+items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[OrderedDict([('', 'Newbie')])],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish'),
                                ('', 'Newbie')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with new empty string key in update args+iter",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[[('', 'Newbie')]],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish'),
                                ('', 'Newbie')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with new empty string key in update kwargs",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[],
            kwargs={'': 'Newbie'},
            exp_obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish'),
                                ('', 'Newbie')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with new non-empty string key in update args+items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[OrderedDict([('New', 'Newbie')])],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish'),
                                ('New', 'Newbie')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with new non-empty string key in update args+iter",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[[('New', 'Newbie')]],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish'),
                                ('New', 'Newbie')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, with new non-empty string key in update kwargs",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[],
            kwargs={'New': 'Newbie'},
            exp_obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish'),
                                ('New', 'Newbie')]),
        ),
        None, None, True
    ),

    # Non-empty NocaseDict, update value of existing key
    (
        "Non-empty dict, updating at existing key in org. case via args+items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[OrderedDict([('Dog', 'Kitten')])],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Kitten'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, updating at existing key in org. case via args+iter",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[[('Dog', 'Kitten')]],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Kitten'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, updating at existing key in org. case via kwargs",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[],
            kwargs={'Dog': 'Kitten'},
            exp_obj=NocaseDict([('Dog', 'Kitten'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, updating at existing key in mixed case via args+items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[OrderedDict([('doG', 'Kitten')])],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Kitten'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, updating at existing key in mixed case via args+iter",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[[('doG', 'Kitten')]],
            kwargs={},
            exp_obj=NocaseDict([('Dog', 'Kitten'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
    (
        "Non-empty dict, updating at existing key in mixed case via kwargs",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            args=[],
            kwargs={'doG': 'Kitten'},
            exp_obj=NocaseDict([('Dog', 'Kitten'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_UPDATE)
@simplified_test_function
def test_NocaseDict_update(testcase,
                           obj, args, kwargs, exp_obj):
    """
    Test function for NocaseDict.update()
    """

    # The code to be tested
    obj.update(*args, **kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseDict features, but that is
    # unavoidable if we want to work through the public interface:

    assert obj == exp_obj  # Uses NocaseDict equality


TESTCASES_NOCASEDICT_CLEAR = [

    # Testcases for NocaseDict.clear()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=NocaseDict(),
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_CLEAR)
@simplified_test_function
def test_NocaseDict_clear(testcase, obj):
    """
    Test function for NocaseDict.clear()
    """

    # The code to be tested
    obj.clear()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseDict features, but that is
    # unavoidable if we want to work through the public interface:

    # The following line uses NocaseDict len
    assert len(obj) == 0  # pylint: disable=len-as-condition


TESTCASES_NOCASEDICT_COPY = [

    # Testcases for NocaseDict.copy()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj: NocaseDict object to be used for the test.
    #   * test_key: Key for testing that copy is a copy, or None to skip.
    #   * test_value: Value for testing that copy is a copy.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dict",
        dict(
            obj=NocaseDict(),
            test_key=None,
            test_value=None,
        ),
        None, None, True
    ),
    (
        "Dict with two items",
        dict(
            obj=NocaseDict([('Dog', 'Cat'), ('Budgie', 'Fish')]),
            test_key='Dog',
            test_value='Kitten',
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_COPY)
@simplified_test_function
def test_NocaseDict_copy(testcase,
                         obj, test_key, test_value):
    """
    Test function for NocaseDict.copy()
    """

    # The code to be tested
    obj_copy = obj.copy()

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    # The verification below also uses some NocaseDict features, but that is
    # unavoidable if we want to work through the public interface:

    assert obj_copy == obj  # Uses NocaseDict equality

    # Verify that the copy is a copy
    if test_key is not None:
        org_value = obj[test_key]  # Uses NocaseDict get
        obj_copy[test_key] = test_value  # Uses NocaseDict set
        now_value = obj[test_key]  # Uses NocaseDict get
        assert now_value == org_value


TESTCASES_NOCASEDICT_EQUAL_HASH = [

    # Testcases for NocaseDict.__hash__(), __eq__(), __ne__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: NocaseDict object #1 to use.
    #   * obj2: NocaseDict object #2 to use.
    #   * exp_obj_equal: Expected equality of the objects.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "Empty dictionary",
        dict(
            obj1=NocaseDict([]),
            obj2=NocaseDict([]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "One item, keys and values equal",
        dict(
            obj1=NocaseDict([('k1', 'v1')]),
            obj2=NocaseDict([('k1', 'v1')]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "One item, keys equal, values different",
        dict(
            obj1=NocaseDict([('k1', 'v1')]),
            obj2=NocaseDict([('k1', 'v1_x')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "One item, keys different, values equal",
        dict(
            obj1=NocaseDict([('k1', 'v1')]),
            obj2=NocaseDict([('k2', 'v1')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "One item, keys equal, values both None",
        dict(
            obj1=NocaseDict([('k1', None)]),
            obj2=NocaseDict([('k1', None)]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "One item, keys different lexical case, values equal",
        dict(
            obj1=NocaseDict([('K1', 'v1')]),
            obj2=NocaseDict([('k1', 'v1')]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Two equal items, in same order",
        dict(
            obj1=NocaseDict([('k1', 'v1'), ('k2', 'v2')]),
            obj2=NocaseDict([('k1', 'v1'), ('k2', 'v2')]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Two items, keys different lexical case, in same order",
        dict(
            obj1=NocaseDict([('K1', 'v1'), ('k2', 'v2')]),
            obj2=NocaseDict([('k1', 'v1'), ('K2', 'v2')]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Two equal items, in different order",
        dict(
            obj1=NocaseDict([('k1', 'v1'), ('k2', 'v2')]),
            obj2=NocaseDict([('k2', 'v2'), ('k1', 'v1')]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Two items, keys different lexical case, in different order",
        dict(
            obj1=NocaseDict([('k1', 'v1'), ('K2', 'v2')]),
            obj2=NocaseDict([('k2', 'v2'), ('K1', 'v1')]),
            exp_obj_equal=True,
        ),
        None, None, True
    ),
    (
        "Comparing unicode value with bytes value",
        dict(
            obj1=NocaseDict([('k1', b'v1')]),
            obj2=NocaseDict([('k2', u'v2')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Matching unicode key with string key",
        dict(
            obj1=NocaseDict([('k1', 'v1')]),
            obj2=NocaseDict([(u'k2', 'v2')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Higher key missing",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Lower key missing",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Dog', 'Cat')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "First non-matching key is less. But longer size!",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Curly', 'Snake'),
                             ('Cozy', 'Dog')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Only non-matching keys that are less. But longer size!",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Alf', 'F'), ('Anton', 'S'), ('Aussie', 'D')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "First non-matching key is greater. But shorter size!",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgio', 'Fish')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Only non-matching keys that are greater. But shorter size!",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Zoe', 'F')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size. First non-matching key is less",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Curly', 'Snake')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size. Only non-matching keys that are less",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Alf', 'F'), ('Anton', 'S')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size. Only non-matching keys that are greater",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Zoe', 'F'), ('Zulu', 'S')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size, only matching keys. First non-matching value is less",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Car')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
    (
        "Same size, only matching keys. First non-matching value is greater",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Caz')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
]

TESTCASES_NOCASEDICT_EQUAL = [

    # Testcases for NocaseDict.__eq__(), __ne__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: NocaseDict object #1 to use.
    #   * obj2: NocaseDict object #2 to use.
    #   * exp_obj_equal: Expected equality of the objects.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "A value raises TypeError when compared (and equal still succeeds)",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', NonComparable()), ('Dog', 'Cat')]),
            exp_obj_equal=False,
        ),
        None, None, True
    ),
]

TESTCASES_NOCASEDICT_HASH = [

    # Testcases for NocaseDict.__hash__() / hash(ncd)

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: NocaseDict object #1 to use.
    #   * obj2: NocaseDict object #2 to use.
    #   * exp_obj_equal: Expected equality of the objects.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    (
        "A value raises TypeError when compared (and hash fails)",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', NonComparable()), ('Dog', 'Cat')]),
            exp_obj_equal=False,
        ),
        TypeError, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_EQUAL_HASH + TESTCASES_NOCASEDICT_EQUAL)
@simplified_test_function
def test_NocaseDict_eq(testcase,
                       obj1, obj2, exp_obj_equal):
    """
    Test function for NocaseDict.__eq__() / ncd1==ncd2
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    eq1 = (obj1 == obj2)
    eq2 = (obj2 == obj1)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert eq1 == exp_obj_equal
    assert eq2 == exp_obj_equal


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_EQUAL_HASH + TESTCASES_NOCASEDICT_EQUAL)
@simplified_test_function
def test_NocaseDict_ne(testcase,
                       obj1, obj2, exp_obj_equal):
    """
    Test function for NocaseDict.__ne__() / ncd1!=ncd2
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    ne1 = (obj1 != obj2)
    ne2 = (obj2 != obj1)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert ne1 != exp_obj_equal
    assert ne2 != exp_obj_equal


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_EQUAL_HASH + TESTCASES_NOCASEDICT_HASH)
@simplified_test_function
def test_NocaseDict_hash(testcase,
                         obj1, obj2, exp_obj_equal):
    """
    Test function for NocaseDict.__hash__() / hash(ncd)
    """

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    hash1 = hash(obj1)
    hash2 = hash(obj2)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    if exp_obj_equal:
        assert hash1 == hash2
    else:
        assert hash1 != hash2


TESTCASES_NOCASEDICT_ORDERING = [

    # Testcases for NocaseDict.__le__(), __lt__(), __ge__(), __gt__() / ord.ops

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * obj1: NocaseDict object #1 to be used.
    #   * obj2: NocaseDict object #2 to be used.
    #   * op: Order comparison operator to be used, as a string (e.g. '>')
    #   * exp_result: Expected result of the comparison, or None.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Empty dicts
    (
        "Empty dicts with >",
        dict(
            obj1=NocaseDict(),
            obj2=NocaseDict(),
            op='>',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Empty dicts with >=",
        dict(
            obj1=NocaseDict(),
            obj2=NocaseDict(),
            op='>=',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Empty dicts with <",
        dict(
            obj1=NocaseDict(),
            obj2=NocaseDict(),
            op='<',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Empty dicts with <=",
        dict(
            obj1=NocaseDict(),
            obj2=NocaseDict(),
            op='<=',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),

    # Equal dicts
    (
        "Equal dicts with >",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='>',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Equal dicts with >=",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='>=',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Equal dicts with <",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='<',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Equal dicts with <=",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='<=',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),

    # Dicts that compare less (obj1 < obj2)
    (
        "Less-comparing dicts with >",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='>',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Less-comparing dicts with >=",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='>=',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Less-comparing dicts with <",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='<',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Less-comparing dicts with <=",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish')]),
            obj2=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            op='<=',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),

    # Dicts that compare greater (obj1 > obj2)
    (
        "Greater-comparing dicts with >",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish')]),
            op='>',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Greater-comparing dicts with >=",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish')]),
            op='>=',
            exp_result=True,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Greater-comparing dicts with <",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish')]),
            op='<',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),
    (
        "Greater-comparing dicts with <=",
        dict(
            obj1=NocaseDict([('Budgie', 'Fish'), ('Dog', 'Cat')]),
            obj2=NocaseDict([('Budgie', 'Fish')]),
            op='<=',
            exp_result=False,
        ),
        TypeError if six.PY3 else None, DeprecationWarning, True
    ),

    # Note: More subtle cases of less- or greater-comparing dicts are not
    # tested because the ordering comparison for NocaseDict is deprecated.
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_NOCASEDICT_ORDERING)
@simplified_test_function
def test_NocaseDict_ordering(testcase,
                             obj1, obj2, op, exp_result):
    """
    Test function for NocaseDict.__le__(), __lt__(), __ge__(), __gt__() / ord.
    """

    comp_str = 'obj1 %s obj2' % op

    # Double check they are different objects
    assert id(obj1) != id(obj2)

    # The code to be tested
    result = eval(comp_str)  # pylint: disable=eval-used

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    assert result == exp_result


def test_unnamed_keys():
    """
    Test function for unnamed keys (key=None). This can be allowed in the
    NocaseDict via an undocumented attribute `allow_unnamed_keys`.
    """

    dic = NocaseDict()
    dic.allow_unnamed_keys = True

    dic[None] = 'a'
    assert None in dic
    assert len(dic) == 1

    a_val = dic[None]
    assert a_val == 'a'

    del dic[None]
    assert None not in dic
    assert not dic



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
Utility functions for pywbem, with no use of other pywbem submodules.
"""

# This module is meant to be safe for 'import *'.

from __future__ import print_function, absolute_import

import re
import inspect
from string import Formatter
import six
try:
    from builtins import ascii
except ImportError:  # py2
    from future_builtins import ascii
try:
    from collections.abc import Mapping, Set, MutableSequence, Sequence
except ImportError:  # py2
    from collections import Mapping, Set, MutableSequence, Sequence

__all__ = []


def _ensure_unicode(obj):
    """
    If the input object is a string, make sure it is returned as a
    :term:`unicode string`, as follows:

    * If the input object already is a :term:`unicode string`, it is returned
      unchanged.
    * If the input string is a :term:`byte string`, it is decoded using UTF-8.
    * Otherwise, the input object was not a string and is returned unchanged.
    """
    if isinstance(obj, six.binary_type):
        return obj.decode("utf-8")
    return obj


def _ensure_bytes(obj):
    """
    If the input object is a string, make sure it is returned as a
    :term:`byte string`, as follows:

    * If the input object already is a :term:`byte string`, it is returned
      unchanged.
    * If the input string is a :term:`unicode string`, it is encoded using
      UTF-8.
    * Otherwise, the input object was not a string and is returned unchanged.
    """
    if isinstance(obj, six.text_type):
        return obj.encode("utf-8")
    return obj


def _ensure_bool(obj):
    """
    If the input object is not `None`, convert it to a :class:`py:bool`.
    If the input object is `None`, return it unchanged.
    """
    if obj is not None:
        obj = bool(obj)
    return obj


def _hash_name(name):
    """
    Hash a CIM name, case-insensitively.

    The name may be `None`.
    """
    if name is None:
        return hash(None)
    return hash(name.lower())


def _hash_item(item):
    """
    Hash an item (CIM value, CIM object), by delegating to its hash function.

    The item may be `None`.
    """
    if isinstance(item, list):
        item = tuple(item)
    return hash(item)


def _hash_dict(dict_):
    """
    Hash a NocaseDict object, by delegating to its hash function.

    The item may be `None`.
    """
    return hash(dict_)


def _stacklevel_above_module(mod_name):
    """
    Return the stack level (with 1 = caller of this function) of the first
    caller that is not defined in the specified module (e.g. "pywbem.cim_obj").

    The returned stack level can be used directly by the caller of this
    function as an argument for the stacklevel parameter of warnings.warn().
    """
    stacklevel = 2  # start with caller of our caller
    frame = inspect.stack()[stacklevel][0]  # stack() level is 0-based
    while True:
        if frame.f_globals.get('__name__', None) != mod_name:
            break
        stacklevel += 1
        frame = frame.f_back
    del frame
    return stacklevel


def _ascii2(value):
    """
    A variant of the `ascii()` built-in function known from Python 3 that:

    (1) ensures ASCII-only output, and
    (2) produces a nicer formatting for use in exception and warning messages
        and other human consumption.

    This function calls `ascii()` and post-processes its output as follows:

    * For unicode strings, a leading 'u' is stripped (u'xxx' becomes 'xxx'),
      if present.

    * For byte strings, a leading 'b' is stripped (b'xxx' becomes 'xxx'),
      if present.

    * For unicode strings, non-ASCII Unicode characters in the range U+0000 to
      U+00FF are represented as '/u00hh' instead of the confusing '/xhh'
      ('/' being a backslash, 'hh' being a 2-digit hex number).

    This function correctly handles values of collection types such as list,
    tuple, dict, and set, by producing the usual Python representation string
    for them. If the type is not the standard Python type (i.e. OrderedDict
    instead of dict), the type name is also shown in the result.

    Returns:
      str: ASCII string
    """

    if isinstance(value, Mapping):
        # NocaseDict in current impl. is not a Mapping; it uses
        # its own repr() implementation (via ascii(), called further down)
        items = [_ascii2(k) + ": " + _ascii2(v)
                 for k, v in six.iteritems(value)]
        item_str = "{" + ", ".join(items) + "}"
        if value.__class__.__name__ == 'dict':
            return item_str
        return "{0}({1})".format(value.__class__.__name__, item_str)

    if isinstance(value, Set):
        items = [_ascii2(v) for v in value]
        item_str = "{" + ", ".join(items) + "}"
        if value.__class__.__name__ == 'set':
            return item_str
        return "{0}({1})".format(value.__class__.__name__, item_str)

    if isinstance(value, MutableSequence):
        items = [_ascii2(v) for v in value]
        item_str = "[" + ", ".join(items) + "]"
        if value.__class__.__name__ == 'list':
            return item_str
        return "{0}({1})".format(value.__class__.__name__, item_str)

    if isinstance(value, Sequence) and \
            not isinstance(value, (six.text_type, six.binary_type)):
        items = [_ascii2(v) for v in value]
        if len(items) == 1:
            item_str = "(" + ", ".join(items) + ",)"
        else:
            item_str = "(" + ", ".join(items) + ")"
        if value.__class__.__name__ == 'tuple':
            return item_str
        return "{0}({1})".format(value.__class__.__name__, item_str)

    if isinstance(value, six.text_type):

        ret = ascii(value)  # returns type str in py2 and py3
        if ret.startswith('u'):
            ret = ret[1:]

        # Convert /xhh into /u00hh.
        # The two look-behind patterns address at least some of the cases that
        # should not be converted: Up to 5 backslashes in repr() result are
        # handled correctly. The failure that happens starting with 6
        # backslashes and even numbers of backslashes above that is not
        # dramatic: The /xhh is converted to /u00hh even though it shouldn't.
        ret = re.sub(r'(?<![^\\]\\)(?<![^\\]\\\\\\)\\x([0-9a-fA-F]{2})',
                     r'\\u00\1', ret)

    elif isinstance(value, six.binary_type):
        ret = ascii(value)  # returns type str in py2 and py3
        if ret.startswith('b'):
            ret = ret[1:]

    elif isinstance(value, (six.integer_types, float)):
        # str() on Python containers calls repr() on the items. PEP 3140
        # that attempted to fix that, has been rejected. See
        # https://www.python.org/dev/peps/pep-3140/.
        # We don't want to make that same mistake, and because ascii() calls
        # repr(), we call str() on the items explicitly. This makes a
        # difference for example for all pywbem.CIMInt values.
        ret = str(value)

    else:
        ret = ascii(value)  # returns type str in py2 and py3

    return ret


class _Ascii2Formatter(Formatter):
    """
    A class derived from `string.Formatter` that supports the conversion
    specifier 'A' to use the `_ascii2()` function (see there for details).

    Note that `string.Formatter` lacks some features of the built-in `format()`
    and `str.format()`` functions. Known are these deficiencies:

    * No unnamed replacements '{}'. Use positional {0} or keyword {f} instead.
    """

    def convert_field(self, value, conversion):
        """
        do any conversion on the resulting object
        """
        if conversion is None:
            return value
        elif conversion == 's':
            return str(value)
        elif conversion == 'r':
            return repr(value)
        elif conversion == 'a':
            return ascii(value)
        elif conversion == 'A':
            return _ascii2(value)
        raise ValueError(
            "Unknown conversion specifier {0!s}".format(conversion))


_ASCII2_FORMATTER = _Ascii2Formatter()


def _format(format_str, *args, **kwargs):
    """
    Return a formatted string, similar to the built-in `format()` function,
    except that it supports the conversion specifier 'A' to use the `_ascii2()`
    function (see there for details).
    """
    return _ASCII2_FORMATTER.format(format_str, *args, **kwargs)


# Pattern for DSP0004 binaryValue; group(1) is value without trailing B
BINARY_VALUE = re.compile(
    r'^([+\-]?(?:[0-1]+))B$',
    flags=(re.UNICODE | re.IGNORECASE))

# Pattern for DSP0004 octalValue
OCTAL_VALUE = re.compile(
    r'^[+\-]?0(?:[1-7]*)$',
    flags=(re.UNICODE))

# Pattern for DSP0004 decimalValue
DECIMAL_VALUE = re.compile(
    r'^[+\-]?(?:0|[1-9][0-9]*)$',
    flags=(re.UNICODE))

# Pattern for DSP0004 hexValue
HEX_VALUE = re.compile(
    r'^[+\-]?0X(?:[0-9A-F]+)$',
    flags=(re.UNICODE | re.IGNORECASE))

# Pattern for DSP0004 realValue (extended by INF, -INF, NAN)
REAL_VALUE = re.compile(
    r'^(?:[+\-]?[0-9]*\.[0-9]+(?:E[+\-]?[0-9]+)?|INF|-INF|NAN)$',
    flags=(re.UNICODE | re.IGNORECASE))


def _integerValue_to_int(value_str):
    """
    Convert a value string that conforms to DSP0004 `integerValue`, into
    the corresponding integer and return it. The returned value has Python
    type `int`, or in Python 2, type `long` if needed.

    Note that DSP0207 and DSP0004 only allow US-ASCII decimal digits. However,
    the Python `int()` function supports all Unicode decimal digits (e.g.
    US-ASCII digits, ARABIC-INDIC digits, superscripts, subscripts) and raises
    `ValueError` for non-decimal digits (e.g. Kharoshthi digits).
    Therefore, the match patterns explicitly check for US-ASCII digits, and
    the `int()` function should never raise `ValueError`.

    Returns `None` if the value string does not conform to `integerValue`.
    """
    m = BINARY_VALUE.match(value_str)
    if m:
        value = int(m.group(1), 2)
    elif OCTAL_VALUE.match(value_str):
        value = int(value_str, 8)
    elif DECIMAL_VALUE.match(value_str):
        value = int(value_str)
    elif HEX_VALUE.match(value_str):
        value = int(value_str, 16)
    else:
        value = None
    return value


def _realValue_to_float(value_str):
    """
    Convert a value string that conforms to DSP0004 `realValue`, into
    the corresponding float and return it.

    The special values 'INF', '-INF', and 'NAN' are supported.

    Note that the Python `float()` function supports a superset of input
    formats compared to the `realValue` definition in DSP0004. For example,
    "1." is allowed for `float()` but not for `realValue`. In addition, it
    has the same support for Unicode decimal digits as `int()`.
    Therefore, the match patterns explicitly check for US-ASCII digits, and
    the `float()` function should never raise `ValueError`.

    Returns None if the value string does not conform to `realValue`.
    """
    if REAL_VALUE.match(value_str):
        value = float(value_str)
    else:
        value = None
    return value



#!/usr/bin/env python
"""
    Test compatibility when changing *args and **kwargs to named args.
"""

import sys


def func_old_args(a, b=None, *args):  # pylint: disable=invalid-name
    """Old function, defined with *args."""

    # We extract a specific optional parameter from *args
    try:
        c = args[0]
    except IndexError:
        c = None

    return a, b, c


def func_old_kwargs(a, b=None, **kwargs):  # pylint: disable=invalid-name
    """Old function, defined with **kwargs."""

    # We extract a specific optional parameter from **kwargs
    try:
        c = kwargs['c']
    except KeyError:
        c = None

    return a, b, c


def func_new(a, b=None, c=None):  # pylint: disable=invalid-name
    """New function, replacing any of the old functions."""

    return a, b, c


def test_func_args(func_args):
    """Examine all possible ways to invoke a function defined with *args."""

    a, b, c = func_args(1)
    assert (a, b, c) == (1, None, None)

    a, b, c = func_args(1, 2)
    assert (a, b, c) == (1, 2, None)

    a, b, c = func_args(1, 2, 3)
    assert (a, b, c) == (1, 2, 3)

    lis = [3, ]
    a, b, c = func_args(1, 2, *lis)
    assert (a, b, c) == (1, 2, 3)

    # The list can even be used for the named args:

    lis = [2, 3]
    a, b, c = func_args(1, *lis)
    assert (a, b, c) == (1, 2, 3)

    lis = [1, 2, 3]
    a, b, c = func_args(*lis)
    assert (a, b, c) == (1, 2, 3)


def test_func_kwargs(func_kwargs):
    """Examine all possible ways to invoke a function defined with **kwargs."""

    a, b, c = func_kwargs(1)
    assert (a, b, c) == (1, None, None)

    a, b, c = func_kwargs(1, 2)
    assert (a, b, c) == (1, 2, None)

    a, b, c = func_kwargs(1, 2, c=3)
    assert (a, b, c) == (1, 2, 3)

    dic = dict(c=3)
    a, b, c = func_kwargs(1, 2, **dic)
    assert (a, b, c) == (1, 2, 3)

    # The dict can even be used for the named args:

    dic = dict(b=2, c=3)
    a, b, c = func_kwargs(1, **dic)
    assert (a, b, c) == (1, 2, 3)

    dic = dict(a=1, b=2, c=3)
    a, b, c = func_kwargs(**dic)
    assert (a, b, c) == (1, 2, 3)


def main():
    """Main function calls the test functs"""

    print("Python version %s" % sys.version)

    print("Testing compatibility for function defined with *args")
    test_func_args(func_old_args)
    test_func_args(func_new)

    print("Testing compatibility for function defined with **kwargs")
    test_func_kwargs(func_old_kwargs)
    test_func_kwargs(func_new)

    print("All tests successful - we can change *args and **kwargs to' \
          ' named args.")
    return 0


if __name__ == '__main__':
    sys.exit(main())
